# -----------------------------------------------------------------------------
# Setup input, run, and read output for AutoAssign, an automatic assignment
# program.  This is done by writing peak lists and spectrum list files needed,
# starting an AutoAssign subprocess, reading AutoAssign peak assignments,
# and applying assignments to peaks.
#
import sys

import os
import socket
import string
import time
import Tkinter
import types

import assigngraph
import atomnames
import axes
import experiments
import pyutil
import sequence
import sparky
import spectrumtype
import sputil
import subprocess
import tkutil

# -----------------------------------------------------------------------------
#
client_executable = 'AsciiClient'
server_executable = 'AutoServer'
server_port = '8050'

# -----------------------------------------------------------------------------
#

# client_assignment_script = 'Execute (Submit Batch)\n'

client_assignment_script = (
  'Execute (Submit Batch)\n' +
  'Execute (Incremental Match, CA,  0.99,   10,   0.005)\n' +
  'Execute (Submit Batch)\n' +
  'Execute (Incremental Match, CB,  0.99,   10,   0.005)\n' +
  'Execute (Submit Batch)\n' +
  'Execute (Incremental Match, HA,  0.99,   10,   0.005)\n' +
  'Execute (Submit Batch)\n' +
  'Execute (Incremental Match, CO,  0.99,   10,   0.005)\n' +
  'Execute (Submit Batch)\n'
  )

# -----------------------------------------------------------------------------
#
class autoassign_dialog(tkutil.Dialog, tkutil.Stoppable):

  def __init__(self):

    self.initialize_spectrum_names()
    self.id_to_peak = {}

    tkutil.Dialog.__init__(self, 'Run AutoAssign')
        
    m = sputil.condition_menu(self.top, 'Molecule: ')
    m.frame.pack(side = 'top', anchor = 'w')
    self.condition_menu = m

    t = tkutil.entry_row(self.top, 'Tolerances (ppm)',
                         ('HN', '.02', 4),
                         ('N15', '.3', 4),
                         ('CA', '.5', 4),
                         ('CB', '.5', 4),
                         ('HA', '.1', 4),
                         ('CO', '.5', 4),
                         )
    t.frame.pack(side = 'top', anchor = 'w')
    self.tolerances = t.entries

    tp = tkutil.file_field(self.top, 'AutoAssign table file: ', 'autoassign')
    tp.frame.pack(side = 'top', anchor = 'w')
    self.table_path_entry = tp
    
    progress_label = Tkinter.Label(self.top, anchor = 'nw', justify = 'left')
    progress_label.pack(side = 'top', anchor = 'w')

    br = tkutil.button_row(self.top,
			   ('Run AutoAssign', self.run_cb),
                           ('Stop', self.stop_cb),
			   ('Close', self.close_cb),
                           ('Help', sputil.help_cb('AutoAssign')),
			   )
    br.frame.pack(side = 'top', anchor = 'w')

    tkutil.Stoppable.__init__(self, progress_label, br.buttons[1])

  # ---------------------------------------------------------------------------
  #
  def initialize_spectrum_names(self):

    self.spectrum_types = (
      'HSQC',
      'HNCO',
      'HNCA',
      'HNcoCA',
      'HNCACB',
      'HNcoCACB',
      'HNHA',
      'HNcoHA',
      )

    self.sparky_type_to_autoassign_type = {
      'n15hsqc':        'HSQC',    
      'hnco':           'HNCO',    
      'hnca':           'HNCA',    
      'hncoca':         'HNcoCA',  
      'hncacb':         'HNCACB',  
      'cbcaconh':       'HNcoCACB',
      'hnha':           'HNHA',    
      'haconh':         'HNcoHA',
      }

    self.type_to_axis_name_table = {
      'HSQC':     {'1H':'HN', '15N':'N15'},
      'HNCO':     {'1H':'HN', '15N':'N15', '13C':'CO'},
      'HNcoCA':   {'1H':'HN', '15N':'N15', '13C':'CA'},
      'HNCA':     {'1H':'HN', '15N':'N15', '13C':'CA'},
      'HNcoCACB': {'1H':'HN', '15N':'N15', '13C':'{CA CB }'},
      'HNCACB':   {'1H':'HN', '15N':'N15', '13C':'{CA CB }'},
      'HNcoHA':   {'1H':('HN','HA'), '15N':'N15'},
      'HNHA':     {'1H':('HN','HA'), '15N':'N15'},
    }

    self.root_spectrum = 'HSQC'
    self.intra_peak_spectra = ('HSQC', 'HNCA', 'HNCACB', 'HNHA',)
    self.inter_peak_spectra = ('HNCO', 'HNcoCA', 'HNCA', 'HNcoCACB',
                               'HNCACB', 'HNcoHA', 'HNHA',)

  # ---------------------------------------------------------------------------
  #
  def run_cb(self):

    condition = sputil.name_to_condition(self.condition_menu.get())
    table_path = self.table_path_entry.get()
    tolerances = self.read_tolerances()

    eset = self.autoassign_experiment_set(condition, tolerances)
    if len(eset.experiments) != 8:
      return

    peak_assignments = self.run_autoassign(eset, tolerances, table_path)

    count = 0
    for alist in peak_assignments.values():
      count = count + len(alist)

    if count == 0:
      self.progress_report('No peak assignments made')
    else:
      self.assign_peaks(peak_assignments)
      self.show_assignments(eset)
      
  # ---------------------------------------------------------------------------
  #
  def read_tolerances(self):

    tolerances = {}
    for e in self.tolerances:
      atom_name = e.label['text']
      tolerance = string.atof(e.variable.get())
      tolerances[atom_name] = tolerance
    tolerances['{CA CB }'] = min(tolerances['CA'], tolerances['CB'])
    return tolerances

  # ---------------------------------------------------------------------------
  # Run AutoAssign and return a table of peak assignments
  #
  def run_autoassign(self, eset, tolerances, table_path):

    self.progress_report('Writing peak lists')
    self.write_autoassign_input_files(eset, tolerances, table_path)

    self.progress_report('Running AutoAssign')
    rshifts = self.autoassign_shifts(eset, table_path)
#    rshifts = self.autoassign_shifts_test(eset, table_path)

    peak_assignments = {}
    for e in eset.experiments:
      self.progress_report('Inferring peak assignments for ' + e.spectrum.name)
      peak_assignments[e] = self.guess_peak_assignments(e, rshifts)

    return peak_assignments

  # ---------------------------------------------------------------------------
  #
  def autoassign_shifts(self, eset, table_path):

    #
    # Start AutoAssign server
    #
    s_proc = subprocess.subprocess(server_executable, server_port)
    s_proc.from_process.close()
    s_proc.to_process.close()

    #
    # Sleep and pray that the server comes up before the client queries it
    #
    time.sleep(1)
    
    #
    # Start AutoAssign ASCII client
    #
    server_host = socket.gethostname()
    aa_proc = subprocess.subprocess(client_executable, server_host,
                                    server_port, '-')

    assignment_script = (
      'Execute (Initialize, %s)\n' % table_path +
      client_assignment_script +
      'Examine (List, Chemical Shifts Publish Format)\n'
      )
    aa_proc.to_process.write(assignment_script)
    aa_proc.to_process.close()
    lines = aa_proc.from_process.readlines()
    code = aa_proc.exit_code()
    if code:
      sys.stderr.write('AutoAssign subprocess exit code = %d\n' % code)
    
    #
    # Kill AutoAssign server
    #
    s_proc.kill()
    code = s_proc.exit_code()
    if code != 9:
      sys.stderr.write('AutoServer subprocess exit code = %d\n' % code)

    cs_lines = self.chemical_shift_lines(lines)
    rshifts = self.resonance_assignments(eset.sequence, cs_lines)
    
    return rshifts
  
  # ---------------------------------------------------------------------------
  #
  def autoassign_shifts_test(self, eset, table_path):

    # For testing w/o AutoAssign
    directory = os.path.split(table_path)[0]
    shifts_path = os.path.join(directory, 'shifts')
    f = open(shifts_path, 'r')
    cs_lines = f.readlines()[4:]
    f.close()

    rshifts = self.resonance_assignments(eset.sequence, cs_lines)
    return rshifts
    
  # ---------------------------------------------------------------------------
  #
  def autoassign_experiment_set(self, condition, tolerances):

    t2s = self.type_to_spectrum_table(condition)
    if t2s == None:
      return []

    elist = []
    for type, spectrum in t2s.items():
      axis_names = self.autoassign_axis_names(spectrum, type)
      tol = pyutil.table_values(tolerances, axis_names)
      self.progress_report('Calculating expected peaks for ' + spectrum.name)
      e = experiments.spectrum_experiment(spectrum, tol)
      e.autoassign_type = type
      elist.append(e)

    return experiments.experiment_set(elist)

  # ---------------------------------------------------------------------------
  #
  def write_autoassign_input_files(self, eset, tolerances, table_path):

    experiments = eset.experiments
    if len(experiments) == 0:
      return

    molecule = experiments[0].spectrum.molecule
    if molecule.name:
      file_prefix = molecule.name + '-'
    else:
      file_prefix = ''

    directory = os.path.split(table_path)[0]
    self.write_peak_lists(eset, directory, file_prefix)
    self.write_spectrum_table(eset, molecule, tolerances,
                              table_path, file_prefix)
    return 1
    
  # ---------------------------------------------------------------------------
  #
  def write_spectrum_table(self, eset, molecule, tolerances,
                           table_path, file_prefix):
    
    f = open(table_path, 'w')

    f.write('#\n' +
            '# %s table file for AutoAssign\n' % molecule.name +
            '#\n')
    f.write('Protein: %s\n' % molecule.name)
    f.write('\n')

    seq = sequence.molecule_sequence(molecule)
    f.write('Sequence: %d %s*\n' % (seq.first_residue_number,
                                    seq.one_letter_codes))
    f.write('\n')

    toltext = ''
    for atom_name in ('HN', 'N15', 'CA', 'CB', 'HA', 'CO'):
      toltext = toltext + ' %s %.3f' % (atom_name, tolerances[atom_name])
    f.write('Tolerances: %s\n' % toltext)
    f.write('\n')

    f.write('Spectra:\n')
    for e in eset.experiments:
      self.write_spectrum_table_entry(e, file_prefix, f)
    f.write('$\n')
    
    f.close()
    
  # ---------------------------------------------------------------------------
  #
  def write_spectrum_table_entry(self, experiment, file_prefix, file):

    type = experiment.autoassign_type
    ref_type = self.root_spectrum
    if type == ref_type:
      ref_type = 'ROOT'

    file_name = file_prefix + type + '.pks'

    intra = type in self.intra_peak_spectra
    inter = type in self.inter_peak_spectra
    
    file.write('%s %s %s %d %d 0 phase: {} {\n' %
               (type, ref_type, file_name, intra, inter))

    tolerances = experiment.tolerances
    spectrum = experiment.spectrum
    min_ppm, max_ppm = spectrum.region
    axis_names = self.autoassign_axis_names(spectrum, type)
    for a in range(spectrum.dimension):
      file.write('{%10s %8.3f %8.3f   0 %8.3f unfolded }\n' %
                 (axis_names[a], min_ppm[a], max_ppm[a], tolerances[a]))
    file.write('}\n')
    
  # ---------------------------------------------------------------------------
  #
  def write_peak_lists(self, eset, directory, file_prefix):

    for e in eset.experiments:
      file_name = file_prefix + e.autoassign_type + '.pks'
      path = os.path.join(directory, file_name)
      self.write_peak_list(e, path)

  # ---------------------------------------------------------------------------
  #
  def write_peak_list(self, experiment, path):

    f = open(path, 'w')

    spectrum = experiment.spectrum
    axes = pyutil.sequence_string(spectrum.nuclei, '%3s ppm  ')
    heading = '# index  %s     intensity  spectrum\n' % axes
    f.write(heading)

    id = 0
    name = self.autoassign_spectrum_name(spectrum)
    for peak in spectrum.peak_list():
      id = id + 1
      ppm = pyutil.sequence_string(peak.frequency, '%8.3f ')
      height = peak.data_height
      line = '%5d %s %12.5g %10s\n' % (id, ppm, height, name)
      f.write(line)
      self.id_to_peak[(name, id)] = peak
      
    f.write('*\n')
    f.close()

  # ---------------------------------------------------------------------------
  #
  def type_to_spectrum_table(self, condition):

    t2slist = {}
    for s in sputil.spectrum_list():
      if s.condition == condition:
        name = self.autoassign_spectrum_name(s)
        if name:
          pyutil.table_push(t2slist, name, s)

    t2s = {}
    error = ''
    for type in self.spectrum_types:
      if t2slist.has_key(type):
        spectra = t2slist[type]
        if len(spectra) == 1:
          t2s[type] = spectra[0]
        else:
          names = reduce(lambda n, s: n + s.name + ' ', spectra, '')
          error = error + '%s can be %s\n' % (type, names)
      else:
        error = error + 'no %s spectrum\n' % type

    if error:
      self.progress_report(error)
      return None

    return t2s

  # ---------------------------------------------------------------------------
  #
  def autoassign_spectrum_name(self, spectrum):

    spectrumtype.guess_spectrum_type(spectrum)
    type = spectrumtype.spectrum_type(spectrum)
    if type and self.sparky_type_to_autoassign_type.has_key(type):
      return self.sparky_type_to_autoassign_type[type]
    return None
    
  # ---------------------------------------------------------------------------
  #
  def autoassign_axis_names(self, spectrum, stype):

    axis_names = []
    n2n = self.type_to_axis_name_table[stype]
    for axis in range(spectrum.dimension):
      nucleus = spectrum.nuclei[axis]
      axis_name = n2n[nucleus]
      if type(axis_name) == types.TupleType:
        heavy_axis = axes.heavy_atom_axes(spectrum)[0]
        attached_axis = axes.attached_proton_axis(spectrum, heavy_axis)
        if axis == attached_axis:
          axis_name = axis_name[0]
        else:
          axis_name = axis_name[1]
      axis_names.append(axis_name)
    return axis_names

  # ---------------------------------------------------------------------------
  # Extract chemical shift lines from output of AutoAssign ascii client.
  #
  def chemical_shift_lines(self, lines):

    start_line = 'AA         HN    N15     CO     CA     CB     HA\n'
    end_line = 'END_MSG\n'

    s = 0
    while s < len(lines) and lines[s] != start_line:
      s = s + 1
    s = s + 2           # skip to first chemical shift line

    e = s
    while e < len(lines) and lines[e] != end_line:
      e = e + 1

    return lines[s:e]
        
  # ---------------------------------------------------------------------------
  # Lines are in AutoAssign publication format.
  #
  def resonance_assignments(self, seq, lines):

    rtable = {}
    for line in lines:
      symbol3 = string.upper(line[0:3])
      symbol = atomnames.aaa_to_a[symbol3]
      number = string.atoi(line[3:7])
      self.set_resonance_shift(symbol, number, 'H', seq, line[7:14], rtable)
      self.set_resonance_shift(symbol, number, 'N', seq, line[14:21], rtable)
      self.set_resonance_shift(symbol, number, 'C', seq, line[21:28], rtable)
      self.set_resonance_shift(symbol, number, 'CA', seq, line[28:35], rtable)
      self.set_resonance_shift(symbol, number, 'CB', seq, line[35:42], rtable)
      self.set_resonance_shift(symbol, number, 'HA', seq, line[42:49], rtable)
      self.set_resonance_shift(symbol, number, 'HA2', seq, line[49:56], rtable)
    return rtable
        
  # ---------------------------------------------------------------------------
  # Add and entry to a resonance shift table is shift_text is non empty
  #
  def set_resonance_shift(self, symbol, number, atom_name, sequence,
                          shift_text, rtable):

    shift = pyutil.string_to_float(shift_text)
    if shift == None:
      return

    r = sequence.resonance(number, symbol, atom_name)
    if r == None:
      return

    rtable[r] = shift
        
  # ---------------------------------------------------------------------------
  # Guess peak assignments from resonance assignments.  AutoAssign can't
  # output peak assignments.  This should be fixed since these guessed peak
  # assignments are not always what AutoAssign actually used.
  #
  def guess_peak_assignments(self, experiment, rshifts):

    alist = []
    for epeak in experiment.expected_peaks:
      if epeak.are_resonances_assigned(rshifts):
        freq = pyutil.table_values(rshifts, epeak.resonances)
        dpeaks = epeak.matching_data_peaks(freq)
        dpeak = sputil.closest_peak(freq, dpeaks, experiment.tolerances)
        if dpeak:
          alist.append((epeak, dpeak))
    return alist
        
  # ---------------------------------------------------------------------------
  #
  def assign_peaks(self, peak_assignments):

    for exper, alist in peak_assignments.items():
      self.progress_report('Assigning peaks for ' + exper.spectrum.name)
      sputil.unassign_spectrum(exper.spectrum)
      for epeak, dpeak in alist:
        epeak.assign(dpeak)
        
  # ---------------------------------------------------------------------------
  #
  def show_assignments(self, eset):

    self.progress_report('Showing peak assignments')
    assigngraph.show_experiment_set(eset)
    self.progress_report('')

# -----------------------------------------------------------------------------
#
def show_dialog():
  tkutil.the_dialog(autoassign_dialog).show_window(1)
