# ----------------------------------------------------------------------------
#
import Tkinter

import atoms
import axes
import sparky
import sputil
import tkutil

class AttachedAtomUnknown(Exception): pass
class AtomUnknown(Exception): pass
class AtomUnassigned(Exception): pass

# -----------------------------------------------------------------------------
#
def extend_noesy_assignment(spectrum, hres):

  if hres and spectrum.dimension == 3:
    try:
      heavy = attached_heavy_atom(hres[0].atom)
      aH, aCN, aH2 = noesy_hsqc_axes(spectrum)
      if atoms.nucleus_type(heavy) == spectrum.nuclei[aCN]:
	return noesy_hsqc_assignment(spectrum,
				     hres[0].atom, hres[1].atom, heavy)
      else:
	return None
    except (AttachedAtomUnknown, AtomUnknown, AtomUnassigned):
      return None
  return hres

# ---------------------------------------------------------------------------
# Find assignments for a given spectrum position where resonances are close.
#
def nearby_noesy_assignments(spectrum, freq, ppm_range):

  c = spectrum.condition
  res0 = c.interval_resonances(freq[0] - ppm_range[0], freq[0] + ppm_range[0])
  res1 = c.interval_resonances(freq[1] - ppm_range[1], freq[1] + ppm_range[1])
  alist = []
  for r0 in res0:
    for r1 in res1:
      alist.append((r0, r1))
  return tuple(alist)

# ---------------------------------------------------------------------------
# Find assignments for a given spectrum position where resonances are close
# and atoms are close.
#
def close_noesy_assignments(spectrum, freq, ppm_range, model, max_distance):

  alist = nearby_noesy_assignments(spectrum, freq, ppm_range)
  clist = []
  for a in alist:
    d = model.distance((a[0].group.number, a[0].atom.name),
                       (a[1].group.number, a[1].atom.name))
    if d != None and d <= max_distance:
      clist.append(a)
  return tuple(clist)

# ---------------------------------------------------------------------------
# Find assignments for a given position in a labelled noesy spectrum where
# resonances are close and the heavy atom is attached to the labelled proton.
#
def nearby_noesy_hsqc_assignments(spectrum, freq, ppm_range):

  aH, aCN, aH2 = noesy_hsqc_axes(spectrum)
  c = spectrum.condition
  rH = c.interval_resonances(freq[aH] - ppm_range[aH],
			     freq[aH] + ppm_range[aH])
  rCN = c.interval_resonances(freq[aCN] - ppm_range[aCN],
			      freq[aCN] + ppm_range[aCN])
  rH2 = c.interval_resonances(freq[aH2] - ppm_range[aH2],
			      freq[aH2] + ppm_range[aH2])

  rHCN = []
  m = spectrum.molecule
  for r in rH:
    hname = atoms.attached_heavy_atom_name(r.atom)
    if hname != None:
      rheavy = c.find_resonance(r.group.name, hname)
      if rheavy != None and rheavy in rCN:
	rHCN.append((r, rheavy))

  alist = []
  for rh, rcn in rHCN:
    for rh2 in rH2:
      a = [None, None, None]
      a[aH] = rh
      a[aCN] = rcn
      a[aH2] = rh2
      alist.append(tuple(a))

  return tuple(alist)

# ---------------------------------------------------------------------------
# Find assignments for a given spectrum position where resonances are close
# and atoms are close.
#
def close_noesy_hsqc_assignments(spectrum, freq, ppm_range,
                                 model, max_distance):

  alist = nearby_noesy_hsqc_assignments(spectrum, freq, ppm_range)
  aH, aCN, aH2 = noesy_hsqc_axes(spectrum)
  clist = []
  for a in alist:
    d = model.distance((a[aH].group.number, a[aH].atom.name),
                       (a[aH2].group.number, a[aH2].atom.name))
    if d != None and d <= max_distance:
      clist.append(a)
  return tuple(clist)

# -----------------------------------------------------------------------------
#
def noesy_hsqc_axes(spectrum):
  if not hasattr(spectrum, 'labelled_proton_axis'):
    identify_hsqc_axes(spectrum)
  aH = spectrum.labelled_proton_axis
  aCN = spectrum.heavy_atom_axis
  aH2 = spectrum.unlabelled_proton_axis
  return aH, aCN, aH2

# -----------------------------------------------------------------------------
#
def is_noesy(spectrum):
  return (spectrum and
	  spectrum.dimension == 2 and
	  len(axes.proton_axes(spectrum)) == 2)

# -----------------------------------------------------------------------------
#
def is_noesy_hsqc(spectrum):
  return (spectrum and
	  spectrum.dimension == 3 and
	  len(axes.proton_axes(spectrum)) == 2)

# ----------------------------------------------------------------------------
#
def atom_pairs_to_hsqc_assignments(spectrum, atom_pairs):
  heavy_nucleus = spectrum.nuclei[spectrum.heavy_atom_axis]
  alist = []
  for h, h2 in atom_pairs:
    try:
      cn = attached_heavy_atom(h)
      if atoms.nucleus_type(cn) == heavy_nucleus:
	alist.append(noesy_hsqc_assignment(spectrum, h, h2, cn))
    except (AtomUnknown, AtomUnassigned, AttachedAtomUnknown):
      pass
  return tuple(alist)
				 
# -----------------------------------------------------------------------------
# Return the assignment associated with atoms for a labelled noesy spectrum
# Throws AtomUnassigned.
#
def noesy_hsqc_assignment(spectrum, labelled_proton,
			  unlabelled_proton, heavy_atom):

  c = spectrum.condition
  res = [None, None, None]
  aH, aCN, aH2 = noesy_hsqc_axes(spectrum)
  res[aH] = atom_resonance(c, labelled_proton)
  res[aH2] = atom_resonance(c, unlabelled_proton)
  res[aCN] = atom_resonance(c, heavy_atom)
  return tuple(res)

# -----------------------------------------------------------------------------
# Throws AtomUnassigned
#
def atom_resonance(condition, atom):
  r = condition.find_resonance(atom)
  if r == None:
    raise AtomUnassigned, atom
  return r

# -----------------------------------------------------------------------------
# Return labelled and unlabelled protons and attached heavy atoms.
# Throws AttachedAtomUnknown and AtomUnknown.
#
def noesy_hsqc_atoms(assignment, spectrum):

  aH, aCN, aH2 = noesy_hsqc_axes(spectrum)
  labelled_proton = assignment[aH].atom
  heavy_atom = assignment[aCN].atom
  unlabelled_proton = assignment[aH2].atom
  other_heavy_atom = attached_heavy_atom(unlabelled_proton)

  return (labelled_proton, heavy_atom, unlabelled_proton, other_heavy_atom)

# -----------------------------------------------------------------------------
# Throws AttachedAtomUnknown and AtomUnknown
#
def attached_heavy_atom(atom):

  heavy_atom_name = atoms.attached_heavy_atom_name(atom)
  if heavy_atom_name == None:
    raise AttachedAtomUnknown, atom

  group_name = atom.group.name
  heavy_atom = atom.molecule.atom(group_name, heavy_atom_name)
  if heavy_atom == None:
    raise AtomUnknown, (group_name, heavy_atom_name)

  return heavy_atom

# ---------------------------------------------------------------------------
#
def noesy_hsqc_spectrum(C13orN15):
  for spectrum in sparky.current_project().spectrum_list():
    if (spectrum.nuclei == (C13orN15, '1H', '1H') or
	spectrum.nuclei == ('1H', C13orN15, '1H') or
	spectrum.nuclei == ('1H', '1H', C13orN15)):
      return spectrum
  return None

# ----------------------------------------------------------------------------
#
def atom_assignment(condition, atoms):
  res = []
  for atom in atoms:
    r = condition.find_resonance(atom)
    if r == None:
      return None
    res.append(r)
  return tuple(res)

# ----------------------------------------------------------------------------
# query user for labelled proton, or auto-determine labelled proton
#
def identify_hsqc_axes(spectrum):

  a1, a2 = axes.proton_axes(spectrum)
  aCN = axes.heavy_atom_axes(spectrum)[0]
  for p in spectrum.peak_list():
    if p.is_assigned:
      res = p.resonances()
      try:
	h1 = attached_heavy_atom(res[a1].atom)
	h2 = attached_heavy_atom(res[a2].atom)
      except (AttachedAtomUnknown, AtomUnknown):
	continue
      if h1 != h2:
	if h1 == res[aCN].atom:
	  set_noesy_hsqc_axes(spectrum, a1, aCN, a2)
	  return
	elif h2 == res[aCN].atom:
	  set_noesy_hsqc_axes(spectrum, a2, aCN, a1)
	  return

  aH = axes.attached_proton_axis(spectrum, aCN)
  if aH == None:
    aH = 2
  aH2 = axes.second_nucleus_axis(spectrum, '1H', aH)
  set_noesy_hsqc_axes(spectrum, aH, aCN, aH2)

# -----------------------------------------------------------------------------
#
def set_noesy_hsqc_axes(spectrum, aH, aCN, aH2):
  spectrum.labelled_proton_axis = aH
  spectrum.unlabelled_proton_axis = aH2
  spectrum.heavy_atom_axis = aCN

# -----------------------------------------------------------------------------
#
def proton_resonances(peak):

  if peak.is_assigned:
    if is_noesy(peak.spectrum):
      return peak.resonances()
    elif is_noesy_hsqc(peak.spectrum):
      aH, aCN, aH2 = noesy_hsqc_axes(peak.spectrum)
      res = peak.resonances()
      return (res[aH], res[aH2])
  return None
