# -----------------------------------------------------------------------------
# Python classes and functions for accessing Sparky data.
#
# The module 'spy' written in C++ defines the classes and functions for
# accessing Sparky data.  This file describes the interface.
#
# The Sparky classes are Project, Spectrum, Peak, Resonance, Label, Line,
# Grid, Molecule, Condition, View.
#
# Here are the data members and methods of these classes:
#
# -----------------------------------------------------------------------------
#
#	class Project:
#	
#	  save_path = '/usr/home/smith/Sparky/Projects/abc.proj'
#	
#	  spectrum_list() = list of <Spectrum instance>
#	  view_list() = list of <View instance>
#	
# -----------------------------------------------------------------------------
#
#	class Spectrum:
#	
#	  name = 'noesy150'			# spectrum name
#	  dimension = 2				# number of axes
#	  data_size = (2048, 4096)		# size along each axis
#	  nuclei = ('1H', '1H')			# '1H', '13C', '15N' for axes
#	  hz_per_ppm = (500.123, 500.123)
#         ppm_per_index = (.00547, .0027)
#	  region = ((.4, -.5), (11.6, 10.7))	# (min-ppm, max-ppm)
#	  noise = 22178.21			# data height noise
#	  condition = <Condition instance>
#	  molecule = <Molecule instance>
#	  data_path = '/data/bigdisk/noesy150'	# nmr data
#	  save_path = '/usr/home/smith/Sparky/Save/noesy150.save'
#         pick_height_thresholds = (-1e5, 1e5)  # negative, positive peaks
#         pick_minimum_linewidth = (.04, .04)   # for peak picking with mouse
#
#	  peak_list()
#	  selected_peaks()
#	  label_list()
#	  line_list()
#	  grid_list()
#	  view_list()
#  	  data_height((2.87, 4.334)) = 321833.23
#	  find_peak(<Resonance list>) = <Peak instance> # can be None
#	  find_peak('G18H8-T17H6') = <Peak instance>	# can be None
#         pick_peaks(((4.3, 2.1), (5.3, 8.4)),          # region
#                    (-100000, 100000),                 # height thresholds
#                    (.04, .04))                        # minimum linewidths
#           = List of <Peak instance>
#         save_value(key, value)                # key = string w/o spaces
#         saved_value(key)                      # value = string w/o newlines
#	
# -----------------------------------------------------------------------------
#
#	class Peak:
#	
#	  __init__(<Spectrum instance>, (4.123, 8.321)) = new <Peak instance>
#
#	  assignment = 'C6H5-C5H1'		# '?-?' for unassigned
#	  is_assigned = 1			# all axes assigned
#	  frequency = (4.35, 5.12)		# in ppm
#	  alias = (0, 0)			# frequency - alias = position
#	  spectrum = <Spectrum instance>
#	  data_height = 104334.23		# closest NMR data value
#	  volume = 1.23e7			# can be None
#	  volume_method = 'gaussian'		# '', 'gaussian',
#						# 'lorentzian', 'box',
#						# 'ellipse', 'peak-group'
#	  volume_error = .15			# error estimate fraction
#	  volume_error_method = 'manual'	# any string
#	  fit_height = 105783.11		# can be None
#	  line_width = (.024, .031)		# ppm, can be None
#	  line_width_method = 'fit'		# any string
#	  note = 'bad overlap'			# settable
#	
#	  selected = 0				# 0 or 1, settable
#	  color = 'white'			# 'red', 'blue', ... settable
#	  label = <Label instance>		# can be None
#
#	  resonances()				# tuple of resonances
#	  peaklets()				# peak group sub-peaks
#	  assign(1, 'C2', "H1'")		# Make assignment for an axis
#	  fit(<View instance>)			# Fit a Gaussian to peak
#	  center()				# Move peak to local maximum
#	  show_assignment_label() = <Label instance>
#	
# -----------------------------------------------------------------------------
#
#	class Resonance:
#	
#	  name = "A17H3'"			# group/atom name
#	  atom = <Atom instance>
#	  group = <Group instance>
#	  frequency = 3.279			# in ppm
#	  deviation = .0143			# one standard deviation (ppm)
#	  condition = <Condition instance>
#         peak_count = 3
#
#	  peak_list()
#	
# -----------------------------------------------------------------------------
#
#	class Label:
#	
#	  text = "T12H6-G13H1'"
#	  shows_assignment = 1			# 0 or 1
#	  selected = 1				# 0 or 1, settable
#	  color = 'green'			# 'red', 'blue', ... settable
#	  peak = <Peak instance>		# attached peak
#	
# -----------------------------------------------------------------------------
#
#	class Line:
#	
#	  __init__(spectrum, start, end)
#	
#	  start = (7.32, 2.43)			# in ppm
#	  end = (7.49, 2.43)			# in ppm
#	  selected = 1				# 0 or 1, settable
#	  color = 'purple'			# 'red', 'blue', ... settable
#
# -----------------------------------------------------------------------------
#
#	class Grid:
#	
#	  position = (5.87, 8.21)		# in ppm
#	  axis = 0				# 0, 1, ..., dim - 1
#	  selected = 0				# 0 or 1, settable
#	  color = 'yellow'			# 'red', 'blue', ... settable
#
# -----------------------------------------------------------------------------
#
#	class Molecule:
#	
#	  name = '24mer'
#	
#	  atom_list()
#	  atom(group_name, atom_name)		# atom object, or None
#	  group_list()
#	  condition_list()
#	  read_atom_positions(pdb_file)		# 1 if succesful, 0 on failure
#         save_value(key, value)                # key = string w/o spaces
#         saved_value(key)                      # value = string w/o newlines
#	
# -----------------------------------------------------------------------------
#
#	class Group:
#	
#	  name = 'T12a'
#	  symbol = 'T'
#	  number = 12				# can be None
#         suffix = 'a'                          # can be empty string
#	  molecule = <Molecule instance>
#	
# -----------------------------------------------------------------------------
#
#	class Atom:
#	
#	  name = 'H3'
#	  group = <Group instance>
#	  molecule = <Molecule instance>
#	  coordinates = (23.4, 13.5, 89.3)	# 3-space coords, or None
#	
# -----------------------------------------------------------------------------
#
#	class Condition:
#	
#	  name = 'ph7 25C'
#	  molecule = <Molecule instance>
#	
#	  resonance_list()
#	  interval_resonances(fmin, fmax)	# tuple of resonances
#	  find_resonance(<Atom instance>)	# returns resonance or None
#	  find_resonance(group_name, atom_name)	# returns resonance or None
#	
# -----------------------------------------------------------------------------
#
#	class View:
#
#         __init__(parent, <Spectrum instance>) # top level window created if
#                                               #   parent is None
#	  name = 'n15noesy'
#	  spectrum = <Spectrum instance>
#	  is_shown = 1				# 0 or 1
#         is_top_level_window = 1               # 0 or 1
#         frame = '.strips.view'                # path of Tk window
#	  center = (3.32, 121.3, 7.12)		# ppm, settable
#	  pixel_size = (.02, .1, .01)		# ppm, settable
#         region = ((3.22, 120.3, 6.12),
#                   (3.42, 122.3, 8.12))
#         visible_depth = (.04, .2, .04)        # z-axis depth for peak display
#         axis_order = (2, 0, 1)                # (x, y, z)
#         contour_levels = <Contour_Levels instance>
#         scale_units = 'ppm'                   # ppm, hz, index
#         show_ornaments = 1
#         show_peaks = 1
#         show_peakgroups = 1
#         show_labels = 1
#         show_lines = 1
#         show_grids = 1
#         subtract_fit_peaks = 0
#         show_scales = 1
#         show_nucleus_names = 0
#         show_scrollbars = 1
#         show_resonance_panels = 0
#         filter_resonances = 0
#         show_slices = 0
#         slice_auto_scale = 0
#         slice_subtract_peaks = 0
#         show_peak_info = 0
#         show_contour_scale = 0
#         show_crosshair = 0
#         show_crosshair_from_other_views = 0
#         show_transpose_crosshair = 0
#         pointer_position = (4.12, 122.2, 6.97) # None if pointer not in view
#
# -----------------------------------------------------------------------------
#
#       class Contour_Levels:
#
#         positive_lowest = 20000               # lowest contour level
#         positive_factor = 1.2                 # step factor between levels
#         positive_levels = 5                   # number of levels
#         positive_color = 'red'                # color of contours
#         negative_lowest = -33000              # lowest contour level
#         negative_factor = 1.3                 # step factor between levels
#         negative_levels = 2                   # number of levels
#         negative_color = 'green'              # color of contours
#
# -----------------------------------------------------------------------------
#
#  Here are some additional Sparky interface functions.
#
#  current_project() = <Project	instance>	# Return current Sparky project
#  command_characters(string)                   # Process key presses
#  installation_path()                          # Sparky installation location
#
#  selected_view() = <View instance>		# can be None
#  selected_peaks()				# Currently selected peaks
#  selected_ornaments()				# Selected peaks, labels, ...
#  unselect_all_ornaments()
#
#  add_selection_callback(func)			# call func with no arguments
#  remove_selection_callback(func)		#  when peak selection changes
#  add_peak_moved_callback(func)		# call func with peak arg
#  remove_peak_moved_callback(func)		#  when peak moved
#
#  initialize_gui(argv, tk)     		# Start the Sparky interface
#  add_menu_command(accel, menu_text, func)	# Add a Sparky command
#  show_manual_url(manual_url)                  # Show manual using Netscape
#  show_peak_list(peak_list)			# Show a peak list window
#  show_resonance_peak_list(res)		# List peaks for resonance
#

import types
from spy import *

# -----------------------------------------------------------------------------
# Specify what object attributes to return when repr(object) called.
#
def object_pretty_print():

  show_attributes(Project, 'save_path')
  show_attributes(Spectrum, 'name', 'dimension', 'data_size', 'nuclei',
                  'region', 'condition', 'molecule', 'data_path', 'save_path')
  show_attributes(Peak, 'assignment', 'frequency', 'spectrum',
                  'data_height', 'volume')
  show_attributes(Resonance, 'name', 'frequency', 'deviation', 'condition')
  show_attributes(Label, 'text', 'selected', 'color')
  show_attributes(Line, 'start', 'end', 'selected', 'color')
  show_attributes(Grid, 'position', 'axis', 'selected', 'color')
  show_attributes(Molecule, 'name')
  show_attributes(Group, 'name', 'number', 'symbol', 'molecule')
  show_attributes(Atom, 'name', 'group', 'molecule', 'coordinates')
  show_attributes(Condition, 'name', 'molecule')
  show_attributes(View, 'name', 'spectrum', 'is_shown', 'center', 'pixel_size')

# -----------------------------------------------------------------------------
# Set an object's __repr__ method to a function showing the specified
# attributes.
#
def show_attributes(c, *attr):

  c.pretty_string_attributes = attr
  c.__dict__['__repr__'] = pretty_instance_string

# -----------------------------------------------------------------------------
# Return a pretty string representation for a class instance
# The pretty_string_attributes member of the instance lists the attributes
# to include in the string.
#
def pretty_instance_string(instance):

  lines = ''
  for attr in instance.pretty_string_attributes:
    value = getattr(instance, attr)
    if type(value) == types.InstanceType:
      classname = value.__class__.__name__
      vstring = '<' + classname
      if hasattr(value, 'name'):
	vstring = vstring + ' ' + value.name
      vstring = vstring + '>'
    else:
      vstring = repr(value)
    lines = lines + '  ' + attr + ' : ' + vstring + '\n'

  classname = instance.__class__.__name__

  return '< ' + classname + '\n' + lines + '>'

# -----------------------------------------------------------------------------
#
object_pretty_print()
